@@ -32,12 +32,16 @@
 #include <config.h>
 #endif
 
-#include <libgnome/libgnome.h>
-#include <libgnomevfs/gnome-vfs.h>
-
 #include <unistd.h>  
 #include <stdio.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/param.h>
+#include <fcntl.h>
+
+#include <libgnome/libgnome.h>
+#include <libgnomevfs/gnome-vfs.h>
 
 #include "gedit-prefs.h"
 #include "gnome-vfs-helpers.h"
@@ -50,6 +54,14 @@
 
 #define NOT_EDITABLE_TAG_NAME "not_editable_tag"
 
+#ifdef MAXPATHLEN
+#define GEDIT_MAX_PATH_LEN  MAXPATHLEN
+#elif defined (PATH_MAX)
+#define GEDIT_MAX_PATH_LEN  PATH_MAX
+#else
+#define GEDIT_MAX_PATH_LEN  2048
+#endif
+
 struct _GeditDocumentPrivate
 {
 	gchar		*uri;
@@ -512,6 +524,22 @@ gedit_document_real_readonly_changed (GeditDocument *document, gboolean readonly
 	g_return_if_fail (document != NULL);
 }
 
+
+gchar*
+gedit_document_get_raw_uri (const GeditDocument* doc)
+{
+	gedit_debug (DEBUG_DOCUMENT, "");
+
+	if (doc->priv->uri == NULL)
+		return NULL;
+	else
+		return g_strdup (doc->priv->uri);
+}
+
+/* 
+ * Returns a well formatted (ready to display) URI in UTF-8 format 
+ * See: gedit_document_get_raw_uri to have a raw uri (non UTF-8)
+ */
 gchar*
 gedit_document_get_uri (const GeditDocument* doc)
 {
@@ -520,7 +548,19 @@ gedit_document_get_uri (const GeditDocument* doc)
 	if (doc->priv->uri == NULL)
 		return g_strdup_printf (_("%s %d"), _("Untitled"), doc->priv->untitled_number);
 	else
-		return gnome_vfs_x_format_uri_for_display (doc->priv->uri);
+	{
+		gchar *res;
+		gchar *uri = g_filename_to_utf8 (doc->priv->uri, -1, NULL, NULL, NULL);
+
+		if (uri == NULL)
+			return g_strdup (_("Invalid file name"));
+		
+		res = gnome_vfs_x_format_uri_for_display (uri);
+
+		g_free (uri);
+
+		return res;
+	}
 }
 
 gchar*
@@ -531,7 +571,19 @@ gedit_document_get_short_name (const GeditDocument* doc)
 	if (doc->priv->uri == NULL)
 		return g_strdup_printf (_("%s %d"), _("Untitled"), doc->priv->untitled_number);
 	else
-		return gnome_vfs_x_uri_get_basename (doc->priv->uri);
+	{
+		gchar *res;
+		gchar *uri = g_filename_to_utf8 (doc->priv->uri, -1, NULL, NULL, NULL);
+		
+		if (uri == NULL)
+			return g_strdup (_("Invalid file name"));
+
+		res = gnome_vfs_x_uri_get_basename (uri);
+
+		g_free (uri);
+
+		return res;
+	}
 }
 
 GQuark 
@@ -914,29 +966,130 @@ gedit_document_save_a_copy_as (GeditDocument* doc, const gchar *uri, GError **er
 	return ret;
 }
 
+#define MAX_LINK_LEVEL 256
+
+/* Does readlink() recursively until we find a real filename. */
+static char *
+follow_symlinks (const gchar *filename, GError **error)
+{
+	gchar *followed_filename;
+	gint link_count;
+
+	g_return_val_if_fail (strlen (filename) + 1 <= GEDIT_MAX_PATH_LEN, NULL);
+
+	followed_filename = g_strdup (filename);
+	link_count = 0;
+
+	while (link_count < MAX_LINK_LEVEL)
+	{
+		struct stat st;
+
+		if (lstat (followed_filename, &st) != 0)
+			/* We could not access the file, so perhaps it does not
+			 * exist.  Return this as a real name so that we can
+			 * attempt to create the file.
+			 */
+			return followed_filename;
+
+		if (S_ISLNK (st.st_mode))
+		{
+			gint len;
+			gchar linkname[GEDIT_MAX_PATH_LEN];
+
+			link_count++;
+
+			len = readlink (followed_filename, linkname, GEDIT_MAX_PATH_LEN - 1);
+
+			if (len == -1)
+			{
+				g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno,
+					     _("Could not read symbolic link information "
+					       "for %s"), followed_filename);
+				g_free (followed_filename);
+				return NULL;
+			}
+
+			linkname[len] = '\0';
+
+			/* If the linkname is not an absolute path name, append
+			 * it to the directory name of the followed filename.  E.g.
+			 * we may have /foo/bar/baz.lnk -> eek.txt, which really
+			 * is /foo/bar/eek.txt.
+			 */
+
+			if (linkname[0] != G_DIR_SEPARATOR)
+			{
+				gchar *slashpos;
+				gchar *tmp;
+
+				slashpos = strrchr (followed_filename, G_DIR_SEPARATOR);
+
+				if (slashpos)
+					*slashpos = '\0';
+				else
+				{
+					tmp = g_strconcat ("./", followed_filename, NULL);
+					g_free (followed_filename);
+					followed_filename = tmp;
+				}
+
+				tmp = g_build_filename (followed_filename, linkname, NULL);
+				g_free (followed_filename);
+				followed_filename = tmp;
+			}
+			else
+			{
+				g_free (followed_filename);
+				followed_filename = g_strdup (linkname);
+			}
+		} else
+			return followed_filename;
+	}
+
+	/* Too many symlinks */
+
+	g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, ELOOP,
+		     _("The file has too many symbolic links."));
+
+	return NULL;
+}
+
 /* FIXME: define new ERROR_CODE and remove the error 
  * strings from here -- Paolo
  */
-#define BAK_EXTENSION_PLUS "~00~"
 
 static gboolean	
 gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
-	       gboolean create_backup_copy, GError **error)
+			     gboolean create_backup_copy, GError **error)
 {
-	FILE  *file;
-	gchar *chars;
-	gchar *fname;
-	gchar *bak_fname = NULL;
-	gchar *temp_fname = NULL;
-	gboolean have_backup = FALSE;
-	gboolean res = FALSE;
-	
+	gchar *filename; /* Filename without URI scheme */
+	gchar *real_filename; /* Final filename with no symlinks */
+	gchar *backup_filename; /* Backup filename, like real_filename.bak */
+	gchar *temp_filename; /* Filename for saving */
+	gchar *slashpos;
+	gchar *dirname;
+	mode_t saved_umask;
+	struct stat st;
+	char *chars;
+	gint chars_len;
+	gint fd;
+	gint retval;
+
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_val_if_fail (doc != NULL, FALSE);
 	g_return_val_if_fail (doc->priv != NULL, FALSE);
 	g_return_val_if_fail (uri != NULL, FALSE);
 
+	retval = FALSE;
+
+	filename = NULL;
+	real_filename = NULL;
+	backup_filename = NULL;
+	temp_filename = NULL;
+
+	/* We don't support non-file:/// stuff */
+
 	if (!gedit_utils_uri_has_file_scheme (uri))
 	{
 		gchar *error_message;
@@ -955,45 +1108,62 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 				_("gedit cannot handle this kind of location in write mode."));
 
 		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, EROFS, error_message);
-		return FALSE;	
+		g_free (error_message);
+		return FALSE;
 	}
-	
-	fname = gnome_vfs_x_format_uri_for_display (uri);
 
-	if (fname != NULL)
+	filename = gnome_vfs_x_format_uri_for_display (uri);
+
+	if (!filename)
 	{
-		temp_fname = g_strconcat (fname, gedit_settings->backup_extension, BAK_EXTENSION_PLUS, NULL);
-		
-		if ((temp_fname == NULL) || (rename (fname, temp_fname) != 0))
-    		{
-      			if (errno != ENOENT)
-			{
-				g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno,
-					_("Could not create the backup copy of the file."));
-		
-				if (fname != NULL) 
-					g_free (fname);
+		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, 0,
+			     _("Invalid filename."));
+		goto out;
+	}
 
-				if (temp_fname != NULL) 
-					g_free (temp_fname);
+	/* Get the real filename and file permissions */
 
-				return FALSE;
+	real_filename = follow_symlinks (filename, error);
 
-			}
-    		}
-  		else
-    			have_backup = TRUE;
-	}	
+	if (!real_filename)
+		goto out;
 
-	if ((fname == NULL) || ((file = fopen (fname, "w")) == NULL))
+	if (stat (real_filename, &st) != 0)
 	{
-		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno,
-			_("Make sure that the path you provided exists, "
-			  "and that you have the appropriate write permissions."));
-						
-		res = FALSE;
+		/* File does not exist? */
+		create_backup_copy = FALSE;
 
-		goto finally;
+		/* Use default permissions */
+		st.st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
+		st.st_uid = getuid ();
+		st.st_gid = getgid ();
+	}
+
+	/* Save to a temporary file.  We set the umask because some (buggy)
+	 * implementations of mkstemp() use permissions 0666 and we want 0600.
+	 */
+
+	slashpos = strrchr (real_filename, G_DIR_SEPARATOR);
+
+	if (slashpos)
+	{
+		dirname = g_strdup (real_filename);
+		dirname[slashpos - real_filename + 1] = '\0';
+	}
+	else
+		dirname = g_strdup (".");
+
+	temp_filename = g_build_filename (dirname, ".gedit-save-XXXXXX", NULL);
+	g_free (dirname);
+
+	saved_umask = umask (0077);
+	fd = g_mkstemp (temp_filename); /* this modifies temp_filename to the used name */
+	umask (saved_umask);
+
+	if (fd == -1)
+	{
+		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno, " ");
+		goto out;
 	}
 
       	chars = gedit_document_get_buffer (doc);
@@ -1003,8 +1173,7 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 		GError *conv_error = NULL;
 		gchar* converted_file_contents = NULL;
 
-		converted_file_contents = g_locale_from_utf8 (chars, -1, NULL, NULL, 
-				&conv_error);
+		converted_file_contents = g_locale_from_utf8 (chars, -1, NULL, NULL, &conv_error);
 
 		if (conv_error != NULL)
 		{
@@ -1017,67 +1186,102 @@ gedit_document_save_as_real (GeditDocument* doc, const gchar *uri,
 			chars = converted_file_contents;
 		}
 	}
-	
-	if (fputs (chars, file) == EOF || fclose (file) == EOF)
+
+	chars_len = strlen (chars);
+	if (write (fd, chars, chars_len) != chars_len)
 	{
-		gchar *errmsg;
+		gchar *msg;
 
 		switch (errno)
 		{
-			case ENOSPC: 
-				errmsg = g_strdup_printf (
-					_("There is no enough disk space to save the file.\n"
-					  "Please, free some space on the disk and try again."));
+			case ENOSPC:
+				msg = _("There is not enough disk space to save the file.\n"
+					"Please free some disk space and try again.");
 				break;
+
 			case EFBIG:
-				errmsg = g_strdup_printf (
-					_("The file is too big."));
+				msg = _("The disk where you are trying to save the file has "
+					"a limitation on file sizes.  Please try saving "
+					"a smaller file or saving it to a disk that does not "
+					"have this limitation.");
 				break;
+
 			default:
-				errmsg = g_strdup_printf (" "); 
+				msg = " ";
+				break;
 		}
 
-		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno, errmsg);
-		g_free (chars);
-
-		res = FALSE;
-
-		goto finally;
+		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno, msg);
+		close (fd);
+		unlink (temp_filename);
+		goto out;
 	}
 
-	g_free (chars);
-
-	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (doc), FALSE);	  
-
-	res = TRUE;
-	
-finally:
-
-	if (have_backup)
+	if (close (fd) != 0)
 	{
-		/* FIXME: should the I/O errors be managed ? - Paolo */
-		if (!res)
-      			rename (temp_fname, fname);
-		else 
-			if (create_backup_copy)
-			{
-				bak_fname = g_strconcat (fname, gedit_settings->backup_extension, NULL);
-				rename (temp_fname, bak_fname);
-			}
-			else
-				unlink (temp_fname);			
+		g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno, " ");
+		unlink (temp_filename);
+		goto out;
 	}
-	
-	if (fname != NULL) 
-		g_free (fname);
 
-	if (temp_fname != NULL) 
-		g_free (temp_fname);
+	/* Move the original file to a backup */
 
-	if (bak_fname != NULL) 
-		g_free (bak_fname);
+	if (create_backup_copy)
+	{
+		gint result;
 
-  	return res;
+		backup_filename = g_strconcat (real_filename, gedit_settings->backup_extension, NULL);
+
+		result = rename (real_filename, backup_filename);
+
+		if (result != 0)
+		{
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, errno,
+				     _("Could not create a backup file."));
+			unlink (temp_filename);
+			goto out;
+		}
+	}
+
+	/* Move the temp file to the original file */
+
+	if (rename (temp_filename, real_filename) != 0)
+	{
+		gint saved_errno;
+
+		saved_errno = errno;
+
+		if (create_backup_copy && rename (backup_filename, real_filename) != 0)
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, saved_errno,
+				     " ");
+		else
+			g_set_error (error, GEDIT_DOCUMENT_IO_ERROR, saved_errno,
+				     " ");
+
+		goto out;
+	}
+
+	/* Restore permissions.  There is not much error checking we can do
+	 * here, I'm afraid.  The final data is saved anyways.
+	 */
+
+	chmod (real_filename, st.st_mode);
+	chown (real_filename, st.st_uid, st.st_gid);
+
+	gtk_text_buffer_set_modified (GTK_TEXT_BUFFER (doc), FALSE);
+
+	retval = TRUE;
+
+	/* Done */
+
+ out:
+
+	g_free (filename);
+	g_free (real_filename);
+	g_free (backup_filename);
+	g_free (temp_filename);
+
+	return retval;
 }
 
 gchar* 
@@ -1483,7 +1687,7 @@ gedit_document_find (GeditDocument* doc, const gchar* str,
 
           	found = gedit_text_iter_forward_search (&iter, str, search_flags,
                                         	&match_start, &match_end,
-                                               	NULL);
+                                               	NULL);	
 		if (found)
 		{
 			gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc),
@@ -1553,6 +1757,12 @@ gedit_document_get_selected_text (GeditDocument *doc, gint *start, gint *end)
 	if (end != NULL)
 		*end = gtk_text_iter_get_offset (&sel_bound); 
 
+	if (gtk_text_iter_equal (&sel_bound, &iter))
+	{
+		gedit_debug (DEBUG_DOCUMENT, "There is no selected text");
+
+		return NULL;
+	}
 	return gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (doc), &iter, &sel_bound, TRUE);
 }
 
@@ -1588,7 +1798,7 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
 	gedit_debug (DEBUG_DOCUMENT, "");
 
 	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
-	g_return_if_fail (replace != NULL);
+	g_return_if_fail (replace != NULL);	
 
 	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
                                     &iter,
@@ -1599,7 +1809,14 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
                                     &sel_bound,
                                     gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
 					                      "selection_bound"));
-		
+
+	if (gtk_text_iter_equal (&sel_bound, &iter))
+	{
+		gedit_debug (DEBUG_DOCUMENT, "There is no selected text");
+
+		return;
+	}
+	
 	gtk_text_iter_order (&sel_bound, &iter);		
 
 	gedit_document_begin_user_action (doc);
@@ -1608,11 +1825,16 @@ gedit_document_replace_selected_text (GeditDocument *doc, const gchar *replace)
 				&iter,
 				&sel_bound);
 
-	gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc),
-				&iter,
-				replace, strlen (replace));
+	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),			
+                                    &iter,
+                                    gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
+					                      "insert"));
+	if (*replace != '\0')
+		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc),
+					&iter,
+					replace, strlen (replace));
 
-	if (doc->priv->last_replace_text)
+	if (doc->priv->last_replace_text != NULL)
 		g_free (doc->priv->last_replace_text);
 
 	doc->priv->last_replace_text = g_strdup (replace);
